from factrainer.base.config import (
    BaseMlModelConfig,
    BasePredictConfig,
    BaseTrainConfig,
)
from factrainer.base.dataset import BaseDataset, Prediction
from factrainer.base.raw_model import RawModel

from .model_container import BaseModelContainer


class SingleModelContainer[
    T: BaseDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](BaseModelContainer[T, U, V, W]):
    def __init__(
        self,
        model_config: BaseMlModelConfig[T, U, V, W],
    ) -> None:
        self._learner = model_config.learner
        self._predictor = model_config.predictor
        self._train_config = model_config.train_config
        self._pred_config = model_config.pred_config

    def train(self, train_dataset: T, val_dataset: T | None = None) -> None:
        self._model = self._learner.train(train_dataset, val_dataset, self.train_config)

    def predict(self, pred_dataset: T) -> Prediction:
        return self._predictor.predict(pred_dataset, self.raw_model, self.pred_config)

    @property
    def raw_model(self) -> U:
        return self._model

    @property
    def train_config(self) -> V:
        return self._train_config

    @train_config.setter
    def train_config(self, config: V) -> None:
        self._train_config = config

    @property
    def pred_config(self) -> W:
        return self._pred_config

    @pred_config.setter
    def pred_config(self, config: W) -> None:
        self._pred_config = config



from .cv.config import PredMode
from .cv.dataset import SplittedDatasetsIndices
from .cv.model_container import CvModelContainer
from .single import SingleModelContainer

__all__ = [
    "PredMode",
    "CvModelContainer",
    "SingleModelContainer",
    "SplittedDatasetsIndices",
]



from enum import Enum, auto
from typing import Any, cast

import numpy as np
import numpy.typing as npt
from factrainer.base.config import (
    BaseLearner,
    BasePredictConfig,
    BasePredictor,
    BaseTrainConfig,
)
from factrainer.base.dataset import BaseDataset, IndexableDataset, Prediction
from factrainer.base.raw_model import RawModel
from joblib import Parallel, delayed

from .dataset import IndexedDatasets
from .raw_model import RawModels


class CvLearner[T: IndexableDataset, U: RawModel, V: BaseTrainConfig](
    BaseLearner[IndexedDatasets[T], RawModels[U], V]
):
    def __init__(self, learner: BaseLearner[T, U, V]) -> None:
        self._learner = learner

    def train(
        self,
        train_dataset: IndexedDatasets[T],
        val_dataset: IndexedDatasets[T] | None,
        config: V,
        n_jobs: int | None = None,
    ) -> RawModels[U]:
        if val_dataset is not None:
            models = Parallel(n_jobs=n_jobs)(
                delayed(self._learner.train)(train.data, val.data, config)
                for train, val in zip(train_dataset.datasets, val_dataset.datasets)
            )
        else:
            models = Parallel(n_jobs=n_jobs)(
                delayed(self._learner.train)(train.data, None, config)
                for train in train_dataset.datasets
            )
        return RawModels(models=models)


class PredMode(Enum):
    OOF_PRED = auto()
    AVG_ENSEMBLE = auto()


class OutOfFoldPredictor[T: IndexableDataset, U: RawModel, W: BasePredictConfig](
    BasePredictor[IndexedDatasets[T], RawModels[U], W]
):
    def __init__(self, predictor: BasePredictor[T, U, W]) -> None:
        self._predictor = predictor

    def predict(
        self,
        dataset: IndexedDatasets[T],
        raw_model: RawModels[U],
        config: W,
        n_jobs: int | None,
    ) -> Prediction:
        y_preds = Parallel(n_jobs=n_jobs)(
            delayed(self._predictor.predict)(_dataset.data, _model, config)
            for _model, _dataset in zip(raw_model.models, dataset.datasets)
        )
        y_oof_pred = self._init_pred(len(dataset), y_preds[0])
        for y_pred, _dataset in zip(y_preds, dataset.datasets):
            y_oof_pred[_dataset.index] = y_pred
        return y_oof_pred

    def _init_pred(self, total_length: int, y_pred: npt.NDArray[Any]) -> Prediction:
        return np.empty(
            tuple([total_length] + list(y_pred.shape[1:])), dtype=y_pred.dtype
        )


class AverageEnsemblePredictor[T: BaseDataset, U: RawModel, W: BasePredictConfig](
    BasePredictor[T, RawModels[U], W]
):
    def __init__(self, predictor: BasePredictor[T, U, W]) -> None:
        self._predictor = predictor

    def predict(
        self,
        dataset: T,
        raw_model: RawModels[U],
        config: W,
        n_jobs: int | None,
    ) -> Prediction:
        y_preds = Parallel(n_jobs=n_jobs)(
            delayed(self._predictor.predict)(dataset, _model, config)
            for _model in raw_model.models
        )
        return cast(Prediction, np.array(y_preds).mean(axis=0))
